/*
 * $Id: formula.c,v 1.13 2000/09/07 07:10:06 keithp Exp $
 *
 * Copyright  2000 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include    "formula.h"
#include    "formulares.h"
#include    <MathLib.h>

#define MAX_VALUES  64

typedef struct _formulaState {
    ULong   currentProgram;
    UInt    currentFunction;
    UInt    valid;
    UInt    topValue;
    UInt    validValue[MAX_VALUES];
    FormNumber	values[MAX_VALUES];
} FormulaState;

typedef struct _formList {
    VoidHand	recHand;
    VoidHand	ptrHand;
    VoidHand	strHand;
    int		n;
    int		cur;
    ULong	current;
    int		str_len;
    FormPtr	form;
    UInt	listID;
    UInt	triggerID;
    Char	**ptr;
    Char	*str;
    Boolean	open;
} FormList;

FormulaState	state;
FormMachine	*machine;
FormPtr		mainForm;
VoidHand	pickleHandle;
FormPickle	*pickle;
DmOpenRef	formulaDB;

FormList	progList, funcList;

UInt
ResizeHandle (VoidHand *hand, ULong size)
{
    if (*hand)
    {
	if (MemHandleResize (*hand, size) != 0)
	{
	    MemHandleFree (*hand);
	    *hand = 0;
	}
    }
    else
    {
	*hand = MemHandleNew (size);
    }
    return *hand != 0;
}

#include <stdarg.h>

void
vsprint (char *dst, char *format, va_list args)
{
    char    c;
    char    *s;
    int	    n;
    
    while ((c = *format++)) {
	switch (c) {
	case '%':
	    if (!(c = *format++))
	    {
		format--;
		break;
	    }
	    switch (c) {
	    case 'd':
		n = va_arg (args, int);
		inttoa (n, dst);
		dst = dst + StrLen (dst);
		break;
	    case 's':
		s = va_arg (args, char *);
		StrCopy (dst, s);
		dst = dst + StrLen (dst);
		break;
	    default:
		*dst++ = c;
		break;
	    }
	    break;
	default:
	    *dst++ = c;
	}
    }
    *dst++ = '\0';
}

FormPtr	errorForm;
Boolean	bitsSaved;

Boolean
ErrorHandleEvent (EventPtr event)
{
    if (errorForm->bitsBehindForm && errorForm->attr.saveBehind)
        bitsSaved = true;
    return false;
}

void
FormError (char *fmt, ...)
{
    char	message[512];
    va_list	args;
    VoidHand	shand = 0;
    VoidPtr	s;
    FieldPtr	field;

    va_start (args, fmt);
    (*vsprint) (message, fmt, args);
    va_end (args);
    shand = MemHandleNew (StrLen (message) + 1);
    if (!shand)
	return;
    errorForm = FrmInitForm (ErrorForm);
    if (!errorForm)
    {
	MemHandleFree (shand);
	return;
    }

    field = FrmGetObjectPtr (errorForm, 
			     FrmGetObjectIndex (errorForm, ErrorField));
    if (field)
    {
	s = MemHandleLock (shand);
	StrCopy (s, message);
	MemHandleUnlock (shand);
	FldSetTextHandle (field, (Handle) shand);
    }
    bitsSaved = false;
    FrmSetEventHandler (errorForm, ErrorHandleEvent);
    (void) FrmDoDialog (errorForm);
    if (!bitsSaved)
	FrmUpdateForm (MainForm, frmRedrawUpdateCode);
    if (field)
    {
	if (shand)
	{
	    FldSetTextHandle (field, 0);
	    MemHandleFree (shand);
	}
    }
    FrmDeleteForm (errorForm);
}

void
InitList (FormList *list)
{
    list->ptrHand = 0;
    list->strHand = 0;
    list->recHand = 0;
    list->n = 0;
    list->str_len = 0;
    list->form = 0;
    list->listID = 0;
    list->triggerID = 0;
    list->cur = noListSelection;
}

int
AddList (FormList *list, ULong r, char *s, int len)
{
    char    **ptr, *str;
    ULong   *rec;
    
    if (!ResizeHandle (&list->ptrHand,
		       (list->n + 1) * sizeof (char *)))
	return 0;
    ptr = MemHandleLock (list->ptrHand);
    ptr[list->n] = (char *) (long) list->str_len;
    MemHandleUnlock (list->ptrHand);
    
    if (!ResizeHandle (&list->strHand,
		       (list->str_len + len + 1)))
	return 0;
    str = MemHandleLock (list->strHand);
    MemMove (str + list->str_len, s, len);
    str[list->str_len+len] = '\0';
    MemHandleUnlock (list->strHand);

    if (!ResizeHandle (&list->recHand, (list->n + 1) * sizeof (ULong)))
	return 0;
    rec = MemHandleLock (list->recHand);
    rec[list->n] = r;
    MemHandleUnlock (list->recHand);
    
    list->n++;
    list->str_len = list->str_len + len + 1;

    return 1;
}

void
SetListCurrent (FormList    *list,
		ULong	    current)
{
    int		r;
    ULong	*rec;
    
    if (list->current != current)
    {
	list->current = current;
	rec = MemHandleLock (list->recHand);
	list->cur = noListSelection;
	for (r = 0; r < list->n; r++)
	    if (rec[r] == current)
	    {
		list->cur = r;
		break;
	    }
    }
}

void
ShowListCurrent (FormList *list)
{
    ListPtr	l;
    ControlPtr  t;
    
    if (list->form)
    {
	l = FrmGetObjectPtr (list->form, 
			     FrmGetObjectIndex (list->form, list->listID));
	LstSetSelection (l, list->cur);
	if (list->triggerID)
	{
	    CharPtr	label;

	    t = FrmGetObjectPtr (list->form, 
				 FrmGetObjectIndex (list->form, list->triggerID));
	    if (0 <= list->cur && list->cur < list->n)
		label = list->ptr[list->cur];
	    else
		label = "";
	    CtlSetLabel (t, label);
	}
    }
}

void
ShowList (FormList *list,
	  FormPtr   form,
	  UInt	    listID,
	  UInt	    triggerID)
{
    ListPtr	l;
    int		n;
    int		r;

    list->form = form;
    list->listID = listID;
    list->triggerID = triggerID;
    if (list->n)
    {
	list->ptr = MemHandleLock (list->ptrHand);
	list->str = MemHandleLock (list->strHand);
	for (r = 0; r < list->n; r++)
	    list->ptr[r] = list->str + (long) list->ptr[r];
    }
    else
    {
	list->ptr = 0;
	list->str = 0;
    }
    l = FrmGetObjectPtr (form, 
			 FrmGetObjectIndex (form, listID));
    LstSetListChoices (l, list->ptr, list->n);
    n = list->n;
    if (n > 10)
	n = 10;
    LstSetHeight (l, n);
    ShowListCurrent (list);
}

void
UnshowList (FormList *list)
{
    if (list->form)
    {
	ListPtr	    l = 0;
	ControlPtr  t = 0;
	int	    r;

	l = FrmGetObjectPtr (list->form, 
			     FrmGetObjectIndex (list->form, list->listID));
	LstSetListChoices (l, 0, 0);
	if (list->triggerID)
	{
	    t = FrmGetObjectPtr (list->form, 
				 FrmGetObjectIndex (list->form, list->triggerID));
	    CtlSetLabel (t, "");
	}
	if (list->n)
	{
	    for (r = 0; r < list->n; r++)
		list->ptr[r] = (Char *) (list->ptr[r] - list->str);
	    MemHandleUnlock (list->ptrHand);
	    MemHandleUnlock (list->strHand);
	}
	list->form = 0;
    }
}

void
FreeList (FormList *list)
{
    UnshowList (list);
    if (list->recHand)
    {
	MemHandleFree (list->recHand);
	list->recHand = 0;
    }
    if (list->ptrHand)
    {
	MemHandleFree (list->ptrHand);
	list->ptrHand = 0;
    }
    if (list->strHand)
    {
	MemHandleFree (list->strHand);
	list->strHand = 0;
    }
    list->n = 0;
    list->str_len = 0;
}

void
ShowProgList (void)
{
    ShowList (&progList, mainForm, MainFormProgList, MainFormProgTrig);
}

void
BuildProgList (void)
{
    UInt	r;
    VoidHand	phand;
    FormPickle	*p;
    int		num;
    char	*name;
    ULong	uniqueID;

    FreeList (&progList);
    num = DmNumRecords (formulaDB);
    for (r = 0; r < num; r++)
    {
	phand = DmQueryRecord (formulaDB, r);
	if (phand == 0)
	    continue;
	if (phand == pickleHandle)
	    p = pickle;
	else
	    p = MemHandleLock (phand);
	if (p->version == FormulaOpVersion)
	{
	    if (DmRecordInfo (formulaDB, r, 0, &uniqueID, 0) == 0)
	    {
		name = PickleName (p);
		AddList (&progList, uniqueID, name, StrLen (name));
	    }
	}
	if (phand != pickleHandle)
	    MemHandleUnlock (phand);
    }
    SetListCurrent (&progList, state.currentProgram);
    if (mainForm)
	ShowProgList ();
}

void
ShowFuncList (void)
{
    ShowList (&funcList, mainForm, MainFormFuncList, MainFormFuncTrig);
}

void
BuildFuncList (void)
{
    UInt	r;
    char	*name;
    FormFunc	*f;

    FreeList (&funcList);
    if (state.valid && pickle && machine)
    {
	for (r = 0; r < machine->exe.numFunc; r++)
	{
	    f = ExeFunc (&machine->exe, r);
	    if (f->label != -2)
	    {
		name = ExeString (&machine->exe, 
				  f->label == -1 ? f->name : f->label);
		AddList (&funcList, (ULong) r, name, StrLen (name));
	    }
	}
    }
    SetListCurrent (&funcList, (ULong) state.currentFunction);
    if (mainForm)
	ShowFuncList ();
}

void
UnloadPickle (void)
{
    if (pickleHandle)
    {
	MemHandleUnlock (pickleHandle);
	pickleHandle = 0;
	pickle = 0;
    }
    if (machine)
    {
	MemPtrFree (machine);
	machine = 0;
    }
}

ULong
AddPickle (FormPickle *new)
{
    ULong	uniqueID;
    UInt	r, num;
    VoidHand	phand;
    FormPickle	*p;
    Int		cmp;

    UnloadPickle ();
    num = DmNumRecords (formulaDB);
    for (r = 0; r < num; r++)
    {
	phand = DmQueryRecord (formulaDB, r);
	if (phand == 0)
	    continue;
	p = MemHandleLock (phand);
	cmp = StrCompare (PickleName (p), PickleName (new));
	MemHandleUnlock (phand);
	if (cmp == 0)
	    break;
    }
    if (r < num)
	phand = DmResizeRecord (formulaDB, r, new->totalSize);
    else
	phand = DmNewRecord (formulaDB, &r, new->totalSize);
    p = MemHandleLock (phand);
    DmWrite (p, 0, new, new->totalSize);
    MemHandleUnlock (phand);
    DmReleaseRecord (formulaDB, r, true);
    DmRecordInfo (formulaDB, r, 0, &uniqueID, 0);
    BuildProgList ();
    return uniqueID;
}

int
GetValue (int	i, FormNumber *n)
{
    FieldPtr	field;
    
    field = FrmGetObjectPtr (mainForm,
			     FrmGetObjectIndex (mainForm,
						MainFormField + i));
    if (!field)
	return 0;
    if (!atodouble (FldGetTextPtr (field), n))
	return 0;
    return 1;
}

void
GetValues (void)
{
    UInt	i;
    int		nused;
    int		nparam;
    FormFunc	*f;
    
    if (!mainForm)
	return;
    nused = 0;
    if (state.valid && machine)
    {
	if (state.currentFunction < machine->exe.numFunc)
	{
	    f = ExeFunc (&machine->exe, state.currentFunction);
	    nparam = f->numParam;
	    if (nparam > MainFormNum)
		nparam = MainFormNum;
	    for (i = 0; i < nparam; i++)
		state.validValue[i] = GetValue (i, &state.values[i]);
	}
    }
}

void
ShowValue (int		i)
{
    FieldPtr	field;
    Char	v[32];

    field = FrmGetObjectPtr (mainForm,
			     FrmGetObjectIndex (mainForm,
						MainFormField + i));
    if (!field)
	return;
    if (state.validValue[i])
    {
	doubletoa (state.values[i], v);
	FldSetSelection (field, 0, FldGetTextLength (field));
	FldInsert (field, v, StrLen (v));
    }
    else
    {
	FldDelete (field, 0, FldGetTextLength (field));
    }
}

void
ShowValues (Boolean doparam)
{
    UInt	i;
    int		nparam, nresult;
    FormFunc	*f;
    
    if (!mainForm)
	return;
    if (!state.valid)
	return;
    if (!machine)
	return;
    if (state.currentFunction >= machine->exe.numFunc)
	return;
    f = ExeFunc (&machine->exe, state.currentFunction);
    nparam = f->numParam;
    if (nparam > MainFormNum)
	nparam = MainFormNum;
    if (f->numResult == -1)
    {
	doparam = true;
	nresult = 0;
    }
    else
    {
	nresult = f->numResult;
	if (nresult + nparam > MainFormNum)
	    nresult = MainFormNum - nparam;
    }
    if (doparam)
	for (i = 0; i < nparam; i++)
	    ShowValue (i);
    for (i = 0; i < nresult; i++)
	ShowValue (i + nparam);
}

void
UpdateLabel (UInt   i,
	     char   *l)
{
    char	label[MaxLabel + 1];

    if (StrLen (l) > MaxLabel)
    {
	MemMove (label, l, MaxLabel);
	label[MaxLabel] = '\0';
	l = label;
    }
    FrmHideObject (mainForm,
		   FrmGetObjectIndex (mainForm, MainFormLabel + i));
    FrmCopyLabel (mainForm, MainFormLabel + i, l);
    FrmShowObject (mainForm,
		   FrmGetObjectIndex (mainForm, MainFormLabel + i));
}

void
UpdateLabels (void)
{
    UInt	i;
    FieldPtr	field;
    FormString	str;
    int		nused;
    int		nparam, nresult, numResult;
    FormFunc	*f;
    FormVar	*v;
    FieldAttrType   attr;
    
    if (!mainForm)
	return;
    nused = 0;
    nparam = 0;
    if (state.valid && machine)
    {
	if (state.currentFunction < machine->exe.numFunc)
	{
	    f = ExeFunc (&machine->exe, state.currentFunction);
	    nparam = f->numParam;
	    if (nparam > MainFormNum)
		nparam = MainFormNum;
	    if (f->numResult == -1)
		numResult = 0;
	    else
		numResult = f->numResult;
	    nresult = numResult;
	    if (nresult + nparam > MainFormNum)
		nresult = MainFormNum - nparam;
	    for (i = 0; i < nparam; i++)
	    {
		v = ExeVar (&machine->exe, f->vars + numResult + i);
		str = v->label;
		if (str == -1)
		    str = v->name;
		UpdateLabel (i, ExeString (&machine->exe, str));
	    }
	    for (i = 0; i < nresult; i++)
	    {
		v = ExeVar (&machine->exe, f->vars + i);
		str = v->label;
		if (str == -1)
		    str = v->name;
		UpdateLabel (i + nparam, ExeString (&machine->exe, str));
	    }
	    nused = nresult + nparam;
	}
    }
    for (i = 0; i < MainFormNum; i++)
    {
	if (i >= nused)
	    UpdateLabel (i, "");
	field = FrmGetObjectPtr (mainForm,
				 FrmGetObjectIndex (mainForm,
						    MainFormField + i));
	if (field)
	{
	    FldEraseField (field);
	    if (i < nused)
	    {
		FldSetUsable (field, true);
		FldGetAttributes (field, &attr);
		if (i < nparam)
		    attr.underlined = grayUnderline;
		else
		    attr.underlined = noUnderline;
		FldSetAttributes (field, &attr);
		FldDrawField (field);
	    }
	    else
	    {
		FldSetUsable (field, false);
	    }
	}
    }
}

Boolean
LoadPickle (void)
{
    UInt	r;
    Err		err;

    UnloadPickle ();
    state.valid = 0;
    err = DmFindRecordByID (formulaDB, state.currentProgram, &r);
    if (err != 0)
	return false;
    pickleHandle = DmQueryRecord (formulaDB, r);
    if (pickleHandle == 0)
	return false;
    machine = MemPtrNew (sizeof (FormMachine));
    if (!machine)
        return false;
    pickle = MemHandleLock (pickleHandle);
    if (FormLoad (pickle, &machine->exe) != 0)
    {
	UnloadPickle ();
	return false;
    }
    state.valid = 1;
    BuildFuncList ();
    MachGlobals (machine);
    return true;
}

FormPickle *(*formParse) (char *, int) = FormParse;

ULong
DoCompile (void)
{
    DmOpenRef	memodb;
    UInt	r, num;
    ULong	memoID;
    VoidHand	rhand;
    ULong	size;
    Byte	*data, *prog;
    UInt	prog_len;
    ULong	*rec;
    FormPtr	form;
    ListPtr	l;
    Word	button;
    Word	selection;
    ULong	uniqueID = 0;
    FormList	list;
    FormPickle	*p;
    
    memodb = DmOpenDatabaseByTypeCreator ('DATA', 'memo', dmModeReadOnly);
    if (memodb == 0)
	return 0;
    num = DmNumRecords (memodb);
    InitList (&list);
    for (r = 0; r < num; r++)
    {
	rhand = DmQueryRecord (memodb, r);
	if (rhand == 0)
	    continue;
	if (DmRecordInfo (memodb, r, 0, &memoID, 0) != 0)
	    continue;
	data = MemHandleLock (rhand);
	size = MemHandleSize (rhand);
	if (*data == '"')
	{
	    --size;
	    ++data;
	    prog = data;
	    prog_len = 0;
	    while (size)
	    {
		if (*data == '"')
		{
		    if (prog_len)
		    {
			if (!AddList (&list, memoID, prog, prog_len))
			    goto bail;
		    }
		    break;
		}
		if (*data == '\n')
		    break;
		--size;
		++data;
		++prog_len;
	    }
	}
	MemHandleUnlock (rhand);
    }
    if (!list.n)
	goto bail;
    
    form = FrmInitForm (CompileForm);
    if (!form)
	goto bail;
    l = FrmGetObjectPtr (form, FrmGetObjectIndex (form, CompileList));
    ShowList (&list, form, CompileList, 0);
    button = FrmDoDialog (form);
    selection = LstGetSelection (l);
    UnshowList (&list);
    FrmDeleteForm (form);
    
    switch (button) {
    case CompileCompile:
	if (selection == noListSelection)
	    break;
	rec = MemHandleLock (list.recHand);
	memoID = rec[selection];
	MemHandleUnlock (list.recHand);
	if (DmFindRecordByID (memodb, memoID, &r) != 0)
	    break;
	rhand = DmQueryRecord (memodb, r);
	if (rhand == 0)
	{
	    FormError ("Can't load indicated memo");
	    break;
	}
	data = MemHandleLock (rhand);
	size = MemHandleSize (rhand);
	DmCloseDatabase (memodb);
	memodb = 0;
	UnloadPickle ();
	FreeList (&list);
	p = (*formParse) (data, size);
	MemHandleUnlock (rhand);
	if (p)
	{
	    uniqueID = AddPickle (p);
	    MemPtrFree (p);
	}
	else
	    LoadPickle ();
	break;
    case CompileCancel:
	break;
    }
    FrmUpdateForm (MainForm, frmRedrawUpdateCode);
bail:
    FreeList (&list);
    if (memodb)
	DmCloseDatabase (memodb);
    return uniqueID;
}

void
SetCursor (void)
{
}

void
SetFunction (UInt function)
{
    int	i;
    state.currentFunction = function;
    for (i = 0; i < MainFormNum; i++)
    {
	state.values[i] = 0.0;
	state.validValue[i] = 0;
    }
    SetListCurrent (&funcList, (ULong) state.currentFunction);
    ShowListCurrent (&funcList);
}

void
SetProgram (ULong uniqueID)
{
    ULong   *rec;
    UInt    f;
    UnloadPickle ();
    state.currentProgram = uniqueID;
    LoadPickle ();
    SetListCurrent (&progList, state.currentProgram);
    ShowListCurrent (&progList);
    if (funcList.recHand)
    {
	rec = MemHandleLock (funcList.recHand);
	f = rec[0];
	MemHandleUnlock (funcList.recHand);
	SetFunction (f);
    }
    if (mainForm)
    {
	UpdateLabels ();
	ShowValues (true);
	FrmUpdateForm (MainForm, frmRedrawUpdateCode);
    }
}
	    
void
RunFunction (void)
{
    FormFunc	*f;
    int		i;
    FormInt	var;
    FormVar	*v;
    int		nparam, nresult;
    int		roffset;

    if (!state.valid || !machine)
    {
	FormError ("No function set loaded");
	return;
    }
    if (state.currentFunction >= machine->exe.numFunc)
    {
	FormError ("Invalid function selected %d", state.currentFunction);
	return;
    }
    f = &machine->exe.funcs[state.currentFunction];
    MachInitialize (machine, f);
    
    GetValues ();
    nparam = f->numParam;
    var = f->vars;
    if (f->numResult != -1)
	var += f->numResult;
    for (i = 0; i < nparam; i++)
    {
	v = ExeVar (&machine->exe, var + i);
	if (i < MainFormNum)
	{
	    if (state.validValue[i])
		MachStore (machine, var + i, state.values[i]);
	    else if (!v->optional)
	    {
		FormError ("Parameter \"%s\" not set",
			   ExeString (&machine->exe,
				      v->label == -1 ? v->name : v->label));
		return;
	    }
	}
    }
    MachRun (machine);
    if (machine->state == FormSuccess)
    {
	var = f->vars;
	if (f->numResult == -1)
	{
	    /*
	     * Outputs equivalent to inputs, refetch params
	     */
	    roffset = 0;
	    nresult = f->numParam;
	}
	else
	{
	    roffset = nparam;
	    nresult = f->numResult;
	}
	for (i = 0; i < nresult; i++)
	{
	    v = ExeVar (&machine->exe, var + i);
	    if (i + roffset < MainFormNum)
	    {
		if (MachIsInit (machine, var + i))
		{
		    state.values[i + roffset] = MachFetch (machine, var + i);
		    state.validValue[i + roffset] = 1;
		}
		else
		    state.validValue[i + roffset] = 0;
	    }
	}
	ShowValues (false);
    }
}

int
SelectProg (UInt selection)
{
    ULong   uniqueID;
    ULong   *rec;
    
    if (selection < 0 || progList.n <= selection)
	return 0;
    rec = MemHandleLock (progList.recHand);
    uniqueID = rec[selection];
    MemHandleUnlock (progList.recHand);
    if (uniqueID != state.currentProgram)
	SetProgram (uniqueID);
    return 1;
}

void
SelectFunc (UInt selection)
{
    int		r;
    ULong    	*rec;

    if (!state.valid)
	return;
    if (!pickle)
	return;
    if (selection < 0 || funcList.n <= selection)
	return;
    rec = MemHandleLock (funcList.recHand);
    r = (int) rec[selection];
    MemHandleUnlock (funcList.recHand);
    if (r != state.currentFunction)
    {
	SetFunction (r);
	UpdateLabels ();
	ShowValues (true);
    }
}

void
DoDelete (void)
{
    UInt	r, num;
    UInt	button, selection;
    VoidHand	phand;
    FormPickle	*p;
    ULong    	*rec;
    char	*name;
    FormPtr	form;
    ListPtr	l;
    UInt	del_cur = 0;
    FormList	list;
    ULong	uniqueID;
    Int		err;

    num = DmNumRecords (formulaDB);
    InitList (&list);
    for (r = 0; r < num; r++)
    {
	phand = DmQueryRecord (formulaDB, r);
	if (phand == 0)
	    continue;
	if (phand == pickleHandle)
	    p = pickle;
	else
	    p = MemHandleLock (phand);
	if (DmRecordInfo (formulaDB, r, 0, &uniqueID, 0) == 0)
	{
	    if (uniqueID == state.currentProgram)
		del_cur = 1;
	    name = PickleName (p);
	    AddList (&list, uniqueID, name, StrLen (name));
	}
	if (phand != pickleHandle)
	    MemHandleUnlock (phand);
    }
    if (!list.n)
	goto bail;
    
    form = FrmInitForm (DeleteForm);

    if (!form)
	goto bail;
    ShowList (&list, form, DeleteList, 0);
    button = FrmDoDialog (form);
    l = FrmGetObjectPtr (form, FrmGetObjectIndex (form, DeleteList));
    selection = LstGetSelection (l);
    UnshowList (&list);
    FrmDeleteForm (form);
    switch (button) {
    case DeleteDelete:
	if (selection == noListSelection)
	    break;
	rec = MemHandleLock (list.recHand);
	uniqueID = rec[selection];
	MemHandleUnlock (list.recHand);
	err = DmFindRecordByID (formulaDB, uniqueID, &r);
	if (err != 0)
	    break;
	phand = DmQueryRecord (formulaDB, r);
	if (phand == pickleHandle)
	{
	    UnloadPickle ();
	}
	DmDeleteRecord (formulaDB, r);
	BuildProgList ();
	if (del_cur)
	{
	    /*
	     * Find another program to load
	     */
	    if (!SelectProg (0))
	    {
		state.valid = 0;
		BuildFuncList ();
		UpdateLabels ();
		ShowValues (true);
	    }
	    UnshowList (&progList);
	    ShowProgList ();
	}
	break;
    case DeleteCancel:
	break;
    }
    FrmUpdateForm (MainForm, frmRedrawUpdateCode);
bail:
    FreeList (&list);
}

/***********************************************************************
 *
 * FUNCTION:    GetFocusObjectPtr
 *
 * DESCRIPTION: P4. This routine returns a pointer to the field object, in 
 *              the current form, that has the focus.
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    pointer to a field object or NULL of there is no focus
 *
 ***********************************************************************/
static FieldPtr GetFocusObjectPtr(void)
{
	FormPtr frm;
	Int focus;
	
	// Get a pointer to tha active form and the index of the form object with focus.
	frm = FrmGetActiveForm();
	focus = FrmGetFocus(frm);
	
	// If no object has the focus return NULL pointer.
	if (focus == noFocus)
		return(NULL);
		
	// Return a pointer to the object with focus.
	return(FrmGetObjectPtr(frm, focus));
}

void
DoCut (void)
{
    FieldPtr	fld = GetFocusObjectPtr ();

    if (fld)
	FldCut(fld);
}

void
DoCopy (void)
{
    FieldPtr	fld = GetFocusObjectPtr ();

    if (fld)
	FldCopy(fld);
}

void
DoPaste (void)
{
    FieldPtr	fld = GetFocusObjectPtr ();

    if (fld)
	FldPaste (fld);
}

void
DoSelect (void)
{
    FieldPtr	fld = GetFocusObjectPtr ();

    if (fld)
	FldSetSelection(fld, 0, FldGetTextLength (fld));
}

void
DoKeyboard (void)
{
    SysKeyboardDialogV10();
}

void
DoGetInfo (void)
{
    FormPtr frm;

    frm = FrmInitForm (InfoForm);
    if (frm)
    {
	FrmDoDialog (frm);
	FrmDeleteForm (frm);
    }
}

Boolean
ApplicationHandleEvent (EventPtr event)
{
    ULong	uniqueID;
    FormPtr	frm;
    
    switch (event->eType) {
    case menuEvent:
	switch (event->data.menu.itemID) {
	case MenuCompile:
	    uniqueID = DoCompile ();
	    if (uniqueID)
		SetProgram (uniqueID);
	    break;
	case MenuDelete:
	    DoDelete ();
	    break;
	case MenuCut:
	    DoCut ();
	    break;
	case MenuCopy:
	    DoCopy ();
	    break;
	case MenuPaste:
	    DoPaste ();
	    break;
	case MenuSelect:
	    DoSelect ();
	    break;
	case MenuKeyboard:
	    DoKeyboard ();
	    break;
	case MenuGetInfo:
	    DoGetInfo ();
	    break;
	}
	break;
    case ctlSelectEvent:
	switch (event->data.ctlSelect.controlID) {
	case MainFormDoIt:
	    RunFunction ();
	    return true;
	}
	break;
    case popSelectEvent:
	switch (event->data.popSelect.listID) {
	case MainFormProgList:
	    SelectProg (event->data.popSelect.selection);
	    break;
	case MainFormFuncList:
	    SelectFunc (event->data.popSelect.selection);
	    break;
	}
	break;
    case frmLoadEvent:
	frm = FrmInitForm(event->data.frmLoad.formID);
	if (event->data.frmLoad.formID == MainForm)
	{
	    mainForm = frm;
	    ShowProgList ();
	    ShowFuncList ();
	    if (state.valid)
	    {
		UpdateLabels ();
		ShowValues (true);
	    }
	}
	FrmSetActiveForm(frm);
	return true;
    case frmOpenEvent:
	FrmDrawForm(FrmGetActiveForm());
	return true;
    case frmUpdateEvent:
	FrmDrawForm(FrmGetActiveForm());
	return true;
    default:
    }
    return false;
}

void
EventLoop (void)
{
    Word	err;
    EventType	e;

    do
    {
	EvtGetEvent(&e, evtWaitForever);

	if (SysHandleEvent(&e)) continue;

	if (MenuHandleEvent((void *)0, &e, &err)) continue;

	if (ApplicationHandleEvent (&e)) continue;

	err = FrmHandleEvent(FrmGetActiveForm(), &e);
    } while (e.eType != appStopEvent);
}

Boolean
StartApplication (void)
{
    Err		err;
    UInt	cardNo;		/* card containing the application database */
    LocalID	dbID;		/* handle for application database */
    UInt	dbAttrs;	/* database attributes */
    
    err = SysLibFind(MathLibName, &MathLibRef);
    if (err)
	err = SysLibLoad(LibType, MathLibCreator, &MathLibRef);
    ErrFatalDisplayIf(err, "Can't find MathLib");
    err = MathLibOpen(MathLibRef, MathLibVersion);
    ErrFatalDisplayIf(err, "Can't open MathLib");
    
    mainForm = 0;
    formulaDB = DmOpenDatabaseByTypeCreator (FormulaDBType, FormulaDBCreator, dmModeReadWrite);
    if (!formulaDB)
    {
	err = DmCreateDatabase (0, "Formulas", FormulaDBCreator, FormulaDBType, false);
	if (err)
	    return false;
	formulaDB = DmOpenDatabaseByTypeCreator (FormulaDBType, FormulaDBCreator, dmModeReadWrite);
	if (!formulaDB)
	    return false;
	err = DmOpenDatabaseInfo (formulaDB, &dbID, NULL, NULL, &cardNo, NULL);
	if (err)
	{
	    DmCloseDatabase (formulaDB);
	    return false;
	}
	dbAttrs = 0;
	err = DmDatabaseInfo (0, dbID, NULL, &dbAttrs, NULL, NULL, NULL, 
			      NULL, NULL, NULL, NULL, NULL, NULL);
	/*
	 * Set the backup flag in the attributes This will ensure that the
	 * database will always get backed upduring HotSync, even if there is
	 * no conduit for this application.
	 */
	dbAttrs = dbAttrs | dmHdrAttrBackup;

	/* Set the new attributes in the database */
	err = DmSetDatabaseInfo(0, dbID, NULL, &dbAttrs, NULL, NULL, NULL, 
				NULL, NULL, NULL, NULL, NULL, NULL);
	if (err)
	{
	    DmCloseDatabase (formulaDB);
	    return false;
	}
    }
    state.currentProgram = 0;
    state.currentFunction = 0;
    InitList (&progList);
    InitList (&funcList);
    if (!PrefGetAppPreferencesV10 (FormulaFileCreator, FormulaVersionNum,
				&state, sizeof (FormulaState)))
    {
	state.valid = 0;
    }
    BuildProgList ();
    if (state.valid)
	LoadPickle ();
    SetCursor ();
    return true;
}
    
void
StopApplication (void)
{
    Err		err;
    UInt	usecount;
    
    GetValues ();
    UnloadPickle ();
    PrefSetAppPreferencesV10 (FormulaFileCreator, FormulaVersionNum,
			   &state, sizeof (FormulaState));
    if (formulaDB)
	DmCloseDatabase (formulaDB);
    err = MathLibClose(MathLibRef, &usecount);
    ErrFatalDisplayIf(err, "Can't close MathLib");
    if (usecount == 0)
	SysLibRemove(MathLibRef);
}

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
#ifdef DEBUG
    set_debug_traps();
    BREAKPOINT();
#endif

    if (cmd == sysAppLaunchCmdNormalLaunch) {
	if (!StartApplication ())
	    return 0;

	FrmGotoForm (MainForm);

	EventLoop ();

	StopApplication ();
    }
    return 0;
}
