/*
 * $Id: lex.c,v 1.11 2000/09/07 09:13:21 keithp Exp $
 *
 * Copyright  2000 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include    "formula.h"
#include    "gram.h"
#include    "MathLib.h"

int	lexline;
int	lexchar;

static unsigned char	*lexin, *lexptr, *lexcurline;
static int	lexlen;

void
lexinit (unsigned char *program, int programSize)
{
    lexin = program;
    lexlen = programSize;
    lexptr = lexin;
    lexcurline = lexptr;
    lexline = 1;
    lexchar = 1;
}

unsigned char *
lexgetline (int *lenp)
{
    unsigned char   *line;
    int		    len;

    line = lexcurline;
    len = 0;
    while (line < lexin + lexlen &&
	   *line != '\0' && *line != '\n')
    {
	len++;
	line++;
    }
    *lenp = len;
    return lexcurline;
}

int
lexc (void)
{
    int	    c;

    if (lexptr == lexin + lexlen)
	c = 0;
    else
	c = *lexptr++;
    ++lexchar;
    if (c == '\n')
    {
	lexcurline = lexptr;
	lexline++;
	lexchar = 1;
    }
    return c;
}

void
unlexc (int c)
{
    if (c != 0)
    {
	if (c == '\n')
	    lexline--;
	else
	    lexchar--;
#ifdef UNIX
	ungetc (c, lexin);
#else
	if (lexptr >= lexin)
	    --lexptr;
#endif
    }
}

static char _keywords[] = 
#define BREAK_STRING	0
    "break\0"
#define CONTINUE_STRING	(BREAK_STRING + 6)
    "continue\0"
#define DECLARE_STRING	(CONTINUE_STRING + 9)
    "declare\0"
#define DEFAULT_STRING	(DECLARE_STRING + 8)
    "default\0"
#define DEFINE_STRING	(DEFAULT_STRING + 8)
    "define\0"
#define DO_STRING	(DEFINE_STRING + 7)
    "do\0"
#define ELSE_STRING	(DO_STRING + 3)
    "else\0"
#define FOR_STRING	(ELSE_STRING + 5)
    "for\0"
#define IF_STRING	(FOR_STRING + 4)
    "if\0"
#define LOCAL_STRING	(IF_STRING + 3)
    "local\0"
#define PRIVATE_STRING	(LOCAL_STRING + 6)
    "private\0"
#define RETURN_STRING	(PRIVATE_STRING + 8)
    "return\0"
#define SWITCH_STRING	(RETURN_STRING + 7)
    "switch\0"
#define WHILE_STRING	(SWITCH_STRING + 7)
    "while\0";

typedef struct _keyword {
    FormInt string;
    FormInt token;
} Keyword;

Keyword	keywords[] = {
{   BREAK_STRING,	BREAK,	},
{   CONTINUE_STRING,	CONTINUE,	},
{   DECLARE_STRING,	DECLARE,	},
{   DEFAULT_STRING,	DEFAULT,	},
{   DEFINE_STRING,	DEFINE,		},
{   DO_STRING,		DO,	},
{   ELSE_STRING,	ELSE,	},
{   FOR_STRING,		FOR,	},
{   IF_STRING,		IF,	},
{   LOCAL_STRING,	LOCAL,	},
{   PRIVATE_STRING,	PRIVATE,	},
{   RETURN_STRING,	RETURN,	},
{   SWITCH_STRING,	SWITCH,	},
{   WHILE_STRING,	WHILE,	},
};

#define NUM_KEYWORDS	((int) (sizeof (keywords) / sizeof (keywords[0])))

int
lookupKeyword (char *s)
{
    int	i;
    
    for	(i = 0; i < NUM_KEYWORDS; i++)
	if (!StrCompare (_keywords + keywords[i].string, s))
	    return keywords[i].token;
    return -1;
}

extern int  yylval;

typedef enum _lexClass {
    LexBlank	= 0x01,
    LexMore	= 0x02, 
    LexLetter	= 0x04,
    LexDigit	= 0x08,
    LexHex	= 0x10,
    LexToken	= 0x20,
    LexString	= 0x40,
    LexPoint	= 0x80,
} LexClass;

static unsigned char	lexMap[128] = {
    0,    		/* ^@ */
    0,			/* ^A */
    0,			/* ^B */
    0,			/* ^C */
    0,			/* ^D */
    0,			/* ^E */
    0,			/* ^F */
    0,			/* ^G */
    0,			/* ^H */
    LexBlank,		/* ^I */
    LexBlank,		/* ^J */
    LexBlank,		/* ^K */
    LexBlank,		/* ^L */
    LexBlank,		/* ^M */
    0,			/* ^N */
    0,			/* ^O */
    0,			/* ^P */
    0,			/* ^Q */
    0,			/* ^R */
    0,			/* ^S */
    0,			/* ^T */
    0,			/* ^U */
    0,			/* ^V */
    0,			/* ^W */
    0,			/* ^X */
    0,			/* ^Y */
    0,			/* ^Z */
    0,			/* ^[ */
    0,			/* ^\ */
    0,			/* ^] */
    0,			/* ^^ */
    0,			/* ^_ */
    LexBlank,		/*    */
    LexToken|LexMore,	/* ! */
    LexString,		/* " */
    0,			/* # */
    0,			/* $ */
    LexToken|LexMore,	/* % */
    LexToken|LexMore,	/* & */
    0,			/* ' */
    LexToken,		/* ( */
    LexToken,		/* ) */
    LexToken|LexMore,	/* * */
    LexToken|LexMore,	/* + */
    LexToken,		/* , */
    LexToken|LexMore,	/* - */
    LexPoint,		/* . */
    LexToken|LexMore,	/* / */
    LexDigit,		/* 0 */
    LexDigit,		/* 1 */
    LexDigit,		/* 2 */
    LexDigit,		/* 3 */
    LexDigit,		/* 4 */
    LexDigit,		/* 5 */
    LexDigit,		/* 6 */
    LexDigit,		/* 7 */
    LexDigit,		/* 8 */
    LexDigit,		/* 9 */
    LexToken,		/* : */
    LexToken,		/* ; */
    LexToken|LexMore,	/* < */
    LexToken|LexMore,	/* = */
    LexToken|LexMore,	/* > */
    LexToken,		/* ? */
    0,			/*  @ */
    LexLetter|LexHex,	/*  A */
    LexLetter|LexHex,	/*  B */
    LexLetter|LexHex,	/*  C */
    LexLetter|LexHex,	/*  D */
    LexLetter|LexHex,	/*  E */
    LexLetter|LexHex,	/*  F */
    LexLetter,		/*  G */
    LexLetter,		/*  H */
    LexLetter,		/*  I */
    LexLetter,		/*  J */
    LexLetter,		/*  K */
    LexLetter,		/*  L */
    LexLetter,		/*  M */
    LexLetter,		/*  N */
    LexLetter,		/*  O */
    LexLetter,		/*  P */
    LexLetter,		/*  Q */
    LexLetter,		/*  R */
    LexLetter,		/*  S */
    LexLetter,		/*  T */
    LexLetter,		/*  U */
    LexLetter,		/*  V */
    LexLetter,		/*  W */
    LexLetter,		/*  X */
    LexLetter,		/*  Y */
    LexLetter,		/*  Z */
    LexToken,		/*  [ */
    0,			/*  \ */
    0,			/*  ] */
    LexToken|LexMore,	/*  ^ */
    LexLetter,		/*  _ */
    0,		    	/*  ` */
    LexLetter|LexHex,	/*  a */
    LexLetter|LexHex,	/*  b */
    LexLetter|LexHex,	/*  c */
    LexLetter|LexHex,	/*  d */
    LexLetter|LexHex,	/*  e */
    LexLetter|LexHex,	/*  f */
    LexLetter,		/*  g */
    LexLetter,		/*  h */
    LexLetter,		/*  i */
    LexLetter,		/*  j */
    LexLetter,		/*  k */
    LexLetter,		/*  l */
    LexLetter,		/*  m */
    LexLetter,		/*  n */
    LexLetter,		/*  o */
    LexLetter,		/*  p */
    LexLetter,		/*  q */
    LexLetter,		/*  r */
    LexLetter,		/*  s */
    LexLetter,		/*  t */
    LexLetter,		/*  u */
    LexLetter,		/*  v */
    LexLetter,		/*  w */
    LexLetter,		/*  x */
    LexLetter,		/*  y */
    LexLetter,		/*  z */
    LexToken,		/*  { */
    LexToken|LexMore,	/*  | */
    LexToken,		/*  } */
    LexToken,		/*  ~ */
};

#define class(c)    (((c) & 0x80) ? 0 : lexMap[c])

FormNumber	yydval;

static const FormNumber hexs[16] = {
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 
};

static FormNumber
char_to_float (unsigned char c)
{
    if ('0' <= c && c <= '9')
	c = c - '0';
    else if ('a' <= c && c <= 'f')
	c = c - 'a' + 10;
    else if ('A' <= c && c <= 'F')
	c = c - 'A' + 10;
    return hexs[c];
}

int
lex (unsigned char *token)
{
    unsigned char   c, cc, nc;
    unsigned char   *t;
    int		    r;

    for (;;)
    {
	c = lexc ();
	r = c;
	while ((cc = class(c)) & LexBlank)
	    c = lexc ();
	if (c == '\0')
	    return 0;
	if (cc & LexToken)
	{
	    if (cc & LexMore)
	    {
		nc = lexc ();
		switch (c) {
		case '!':
		    if (nc == '=')
			return NE;
		    break;
		case '%':
		    if (nc == '=')
			return MODEQ;
		    break;
		case '&':
		    if (nc == '&')
			return AND;
		    if (nc == '=')
			return LANDEQ;
		    break;
		case '*':
		    if (nc == '=')
			return TIMESEQ;
		    break;
		case '+':
		    if (nc == '+')
			return PLUSPLUS;
		    if (nc == '=')
			return PLUSEQ;
		    break;
		case '-':
		    if (nc == '-')
			return MINUSMINUS;
		    if (nc == '=')
			return MINUSEQ;
		    break;
		case '/':
		    if (nc == '/')
		    {
			nc = lexc ();
			if (nc == '=')
			    return DIVEQ;
			unlexc (nc);
			return DIV;
		    }
		    if (nc == '=')
			return DIVIDEEQ;
		    break;
		case '<':
		    if (nc == '=')
			return LE;
		    break;
		case '=':
		    if (nc == '=')
			return EQ;
		    break;
		case '>':
		    if (nc == '=')
			return GE;
		    break;
		case '^':
		    if (nc == '^')
		    {
			nc = lexc ();
			if (nc == '=')
			    return LXOREQ;
			unlexc (nc);
			return LXOR;
		    }
		    if (nc == '=')
			return POWEQ;
		    break;
		case '|':
		    if (nc == '|')
			return OR;
		    if (nc == '=')
			return LOREQ;
		    break;
		}
		unlexc (nc);
	    }
	    return c;
	}
	if (cc & LexLetter)
	{
	    t = token;
	    while (t < token + LEX_TOKEN - 1)
	    {
		*t++ = c;
		c = lexc ();
		if (!((class(c) & (LexLetter|LexDigit))))
		{
		    unlexc (c);
		    break;
		}
	    }
	    *t = '\0';
	    r = lookupKeyword (token);
	    if (r != -1)
		return r;
	    return NAME;
	}
	if (cc & (LexDigit|LexPoint))
	{
	    FormNumber	n = 0;
	    FormNumber	f = 0;
	    FormNumber	e = 1;
	    FormNumber	base = Number(10.0);
	    FormInt		negative;
	    unsigned char	nc;
    
	    t = token;
	    if (c == '0')
	    {
		nc = lexc ();
		if (nc == 'x' || nc == 'X')
		{
		    base = Number(16.0);
		    c = lexc ();
		}
		else if (nc == 'b' || nc == 'B')
		{
		    base = Number(2.0);
		    c = lexc ();
		}
		else if (nc == 'o' || nc == 'O')
		{
		    base = Number(8.0);
		    c = lexc ();
		}
		else if (nc == 't' || nc == 'T')
		{
		    base = Number(10.0);
		    c = lexc ();
		}
		else if (nc == '.')
		{
		    unlexc (nc);
		}
		else if (!(class (nc) & LexDigit))
		{
		    unlexc (nc);
		}
		else
		{
		    base = Number(8.0);
		    c = nc;
		}
	    }
	    while (class (c) & (LexDigit | (base == Number(16.0) ? LexHex : 0)))
	    {
		n = n * base + char_to_float (c);
		c = lexc ();
	    }
	    if (c == '.')
	    {
		c = lexc ();
		f = base;
		while (class (c) & (LexDigit | (base == Number(16.0) ? LexHex : 0)))
		{
		    n = n + char_to_float (c) / f;
		    f = f * base;
		    c = lexc ();
		}
	    }
	    if (c == 'e' || c == 'E')
	    {
		c = lexc ();
		negative = 0;
		if (c == '-')
		{
		    negative = 1;
		    c = lexc ();
		} 
		else if (c == '+')
		{
		    c = lexc ();
		}
		e = NumberZero;
		while (class (c) & (LexDigit | (base == Number(16.0) ? LexHex : 0)))
		{
		    e = e * base + char_to_float (c);
		    c = lexc ();
		}
		if (negative)
		    e = -e;
		n = n * FormPow (base, e);
	    }
	    unlexc (c);
	    yydval = n;
	    return NUMBER;
	}
	if (cc & LexString)
	{
	    c = lexc ();
	    t = token;
	    while (t < token + LEX_TOKEN - 1)
	    {
		if (c == '\\')
		{
		    c = lexc ();
		    if (class (c) & LexDigit)
		    {
			FormInt i;
			i = c - '0';
			c = lexc ();
			if (class (c) & LexDigit)
			{
			    i = i * 8 + c - '0';
			    c = lexc ();
			    if (class (c) & LexDigit)
				i = i * 8 + c - '0';
			    else
				unlexc (c);
			}
			else
			    unlexc (c);
			c = i;
		    }
		    else
		    {
			switch (c) {
			case 'a':
			    c = '\a';
			    break;
			case 'b':
			    c = '\b';
			    break;
			case 'f':
			    c = '\f';
			    break;
			case 'n':
			    c = '\n';
			    break;
			case 'r':
			    c = '\r';
			    break;
			case 'v':
			    c = '\v';
			    break;
			}
		    }
		}
		*t++ = c;
		c = lexc ();
		if (c == '\0' || (class (c) & LexString))
		    break;
	    }
	    *t = '\0';
	    return LABEL;
	}
	
	if (c == '#')
	{
	    do
		c = lexc ();
	    while (c && c != '\n');
	}
	else
	    return ERROR;
    }
}

int
atodouble (char *s, FormNumber *n)
{
    int	    neg;
    int	    l;
    unsigned char   token[LEX_TOKEN];
    
    if (!s)
	return 0;
    lexinit (s, StrLen (s));
    neg = 0;
    l = lex (token);
    if (l == '-')
    {
	neg = 1;
	l = lex (token);
    } 
    else if (l == '+')
    {
	l = lex (token);
    }
    if (l != NUMBER)
	return NumberZero;
    if (neg)
	yydval = -yydval;
    *n = yydval;
    return 1;
}

void
udtoa(double d, char *s, int p)
{
    char    buf[16], *b;
    double  i;
    int	    t;

    t = p;
    b = buf + 16;
    *--b = '\0';
    do
    {
	i = fmod (d, 10.0);
	d = d / 10.0;
	if (!t || i >= 1.0)
	{
	    *--b = '0' + i;
	    t = 0;
	}
	p--;
    } while (d >= 1.0);
    while (p-- > 0)
	*--b = '0';
    while ((*s++ = *b++));
}

void
ultoa (unsigned long l, char *s, int p)
{
    char    buf[16], *b;
    unsigned long   d;
    int	    t;

    t = p;
    b = buf + 16;
    *--b = '\0';
    do
    {
	d = l % 10;
	l = l / 10;
	if (!t || d)
	{
	    *--b = '0' + d;
	    t = 0;
	}
	p--;
    } while (l);
    if (b != buf+15)
	while (p-- > 0)
	    *--b = '0';
    while ((*s++ = *b++));
}

void
inttoa (int i, char *s)
{
    if (i < 0)
    {
	*s++ = '-';
	i = -i;
    }
    ultoa ((unsigned long) i, s, 0);
}

#define BIG	1e9
#define SMALL	1e-4
#define EFRAC	1e10
#define NEFRAC	10
#define FRAC	1e14
#define NFRAC	14
#define eFRAC	1e9
#define NeFRAC	9

void
doubletoa (FormNumber n, char *s)
{
    FormNumber	    e, ip;
    int		    exp;
    FormNumber	    frac = FRAC;
    int		    nfrac = NFRAC;
    
    if (n < NumberZero)
    {
	*s++ = '-';
	n = -n;
    }
    exp = 0;
    if (n == NumberZero)
    {
	*s++ = '0';
	*s++ = '\0';
    }
    else
    {
	e = Number(1.0);
	exp = 0;
	if (n >= 1.0)
	{
	    while (n / e >= Number(10.0))
	    {
		exp++;
		e *= Number (10.0);
	    }
	    if (n >= BIG)
	    {
		frac = EFRAC;
		nfrac = NEFRAC;
		n = n / e;
	    }
	    else
	    {
		frac = FRAC/e;
		nfrac = NFRAC-exp;
		exp = 0;
	    }
	}
	else if (n < SMALL)
	{
	    while (n * e < Number(1.0))
	    {
		exp--;
		e *= Number(10.0);
	    }
	    n = n * e;
	    frac = eFRAC;
	    nfrac = NeFRAC;
	}
	ip = floor (n);
	udtoa (ip, s, 0);
	s += StrLen(s);
	n = n - ip;
	n = n * frac + 0.5;
	if (n >= 1.0)
	{
	    *s++ = '.';
	    udtoa (n, s, nfrac);
	    s += StrLen(s);
	}
	if (exp)
	{
	    *s++ = 'e';
	    if (exp < 0)
	    {
		*s++ = '-';
		exp = -exp;
	    }
	    ultoa (exp, s, 0);
	}
    }
}
